/*----------------------------------------------------------------------
  File    : cm4seqs.c
  Contents: prefix tree management for closed and maximal sequences
  Author  : Christian Borgelt
  History : 2014.07.16 file created
            2014.07.17 closed/maximal filtering and reporting debugged
            2014.07.23 hierarchy of conditional repositories added
            2016.10.18 extended to item sequences (with repetition)
            2016.10.24 unnecessary functions removed (projections)
            2016.10.25 projection to extension item fixed (cms_project)
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>
#include "cm4seqs.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Prefix Tree Functions
----------------------------------------------------------------------*/

CMSTREE* cms_createx (MEMSYS *mem, ITEM size, int target, int dir)
{                               /* --- create a c/m prefix tree */
  CMSTREE *cms;                 /* created c/m sequence prefix tree */

  assert(size > 0);             /* check the function arguments */
  cms = (CMSTREE*)calloc(1, sizeof(CMSTREE)
                         +(size_t)(size-1) *sizeof(ITEM));
  if (!cms) return NULL;        /* create a c/m sequence prefix tree */
  cms->target = target;         /* and initialize its fields */
  cms->dir    = (dir < 0) ? -1 : +1;
  cms->size   = size;
  cms->item   = -1;             /* clear the tree's prefix item */
  cms->mem    = (mem) ? mem : ms_create(sizeof(CMSNODE), 4095);
  if (!cms->mem) { free(cms); return NULL; }
  while (--size >= 0)           /* traverse the items and */
    cms->pos[size] = -1;        /* clear the item positions */
  cms->root.supp    =  0;       /* initialize the root node */
  cms->root.item    = -1;       /* (no item, no other nodes) */
  cms->root.height  =  0;
  cms->root.sibling = cms->root.children = NULL;
  return cms;                   /* return the created prefix tree */
}  /* cms_createx() */

/*--------------------------------------------------------------------*/

void cms_clear (CMSTREE *cms)
{                               /* --- clear a c/m prefix tree */
  assert(cms);                  /* check the function argument */
  ms_clear(cms->mem, 0);        /* clear memory management system */
  cms->item = -1;               /* clear the tree's prefix item */
  cms->root.supp    = 0;
  cms->root.height  = 0;        /* reinitialize the root node */
  cms->root.sibling = cms->root.children = NULL;
}  /* cms_clear() */

/*--------------------------------------------------------------------*/

void cms_delete (CMSTREE *cms, int delms)
{                               /* --- delete a c/m prefix tree */
  assert(cms);                  /* check the function arguments */
  if      (delms >  0) ms_delete(cms->mem);
  else if (delms >= 0) cms_clear(cms);
  free(cms);                    /* delete memory system or nodes */
}  /* cms_delete() */           /* and delete the base structure */

/*--------------------------------------------------------------------*/

int cms_add (CMSTREE *cms, const ITEM *items, ITEM n, RSUPP supp)
{                               /* --- add a sequence to a c/m tree */
  ITEM    i;                    /* buffer for an item */
  CMSNODE **p;                  /* pointer to insertion position */
  CMSNODE *node;                /* to insert new nodes */

  assert(cms                    /* check the function arguments */
  &&    (items || (n <= 0)) && (supp >= 0));
  node = &cms->root; n += 1;    /* start at the root node */
  do {                          /* traverse the items of the sequence */
    if (supp > node->supp)      /* adapt the node support */
      node->supp = supp;        /* (root represents empty sequnece) */
    if (n    > node->height)    /* adapt the subtree height */
      node->height = n;         /* (may be increased by addition) */
    if (--n <= 0) return 0;     /* if all items are processed, abort */
    i = *items++;               /* get the next item in the seq. and */
    p = &node->children;        /* traverse the list of children */
    if (cms->dir < 0) while (*p && ((*p)->item > i)) p = &(*p)->sibling;
    else              while (*p && ((*p)->item < i)) p = &(*p)->sibling;
    node = *p;                  /* find the item/insertion position */
  } while (node && (node->item == i));
  node = (CMSNODE*)ms_alloc(cms->mem);
  if (!node) return -1;         /* create a new prefix tree node */
  node->supp    = supp;         /* store the support of the sequence */
  node->item    = i;            /* and the current/last item */
  node->height  = n;            /* store the number of rem. items, */
  node->sibling = *p;           /* insert the created node */
  *p = node;                    /* into the sibling list */
  while (--n > 0) {             /* traverse the remaining items */
    node = node->children = (CMSNODE*)ms_alloc(cms->mem);
    if (!node) return -1;       /* create a new prefix tree node */
    node->supp    = supp;       /* store support of the sequence */
    node->item    = *items++;   /* and the current/last item */
    node->height  = n;          /* store the number of rem. items, */
    node->sibling = NULL;       /* there are no siblings yet */
  }
  node->children = NULL;        /* last created node is a leaf */
  return 0;                     /* return 'ok' */
}  /* cms_add() */

/*--------------------------------------------------------------------*/

RSUPP cms_get (CMSTREE *cms, const ITEM *items, ITEM n)
{                               /* --- get support of an item seq. */
  ITEM    i;                    /* buffer for an item */
  CMSNODE *p;                   /* to traverse the nodes */

  assert(cms && (items || (n <= 0))); /* check function arguments */
  p = &cms->root;               /* start search at the root node */
  while (n > 0) {               /* traverse the items of the seq. */
    if (n > p->height)          /* check whether subtree is high */
      return -1;                /* enough, otherwise abort */
    i = *items++; n--;          /* try to find a child node */
    p = p->children;            /* with the next item in the seq. */
    if (cms->dir < 0) while (p && (p->item > i)) p = p->sibling;
    else              while (p && (p->item < i)) p = p->sibling;
    if (!p || (p->item != i))   /* if a node with the next item */
      return -1;                /* does not exist in the tree, */
  }                             /* abort the search with failure */
  return p->supp;               /* return support of the item seq. */
}  /* cms_get() */

/*--------------------------------------------------------------------*/

static int super (CMSNODE *node, const ITEM *items, ITEM n, SUPP supp)
{                               /* --- check for a super-sequence */
  assert(items && (n > 0) && (supp > 0)); /* check arguments */
  for (node = node->children; node; node = node->sibling) {
    if (node->supp < supp)      /* traverse the child node list, but */
      continue;                 /* skip children with insuff. support */
    if (node->item == *items) { /* if at node with matching item */
      if (n <= 1) return -1;    /* if last item was matched, abort */
      if ((node->height >= n)   /* if subtree is high enough */
      &&  super(node, items+1, n-1, supp))
        return -1; }            /* check children for rem. items */
    else {                      /* if at node with different item */
      if ((node->height >  n)   /* if subtree is high enough */
      &&  super(node, items,   n,   supp))
        return -1;              /* check children for rem. items */
    }                           /* and abort if a superset is found */
  }
  return 0;                     /* return 'no superset exists' */
}  /* super() */

/*--------------------------------------------------------------------*/

int cms_super (CMSTREE *cms, const ITEM *items, ITEM n, RSUPP supp)
{                               /* --- check support of a sequence */
  assert(cms                    /* check the function arguments */
  &&    (items || (n <= 0)) && (supp > 0));
  if (n <= 0)                   /* if no item, check root support */
    return (cms->root.supp >= supp) ? -1 : 0;
  return super(&cms->root, items, n, supp);
}  /* cms_super() */            /* recursively check for super-seq. */

/*--------------------------------------------------------------------*/

static void delete (CMSNODE *node, MEMSYS *mem)
{                               /* --- delete a c/m subtree */
  CMSNODE *c, *d;               /* to traverse the child nodes */

  assert(node && mem);          /* check the function arguments */
  for (c = node->children; c;){ /* traverse the list of children */
    d = c; c = d->sibling;      /* note the current child node, */
    delete(d, mem);             /* go to its right sibling and */
  }                             /* finally delete the child node */
  ms_free(mem, node);           /* delete the node itself */
}  /* delete() */

/*--------------------------------------------------------------------*/

static CMSNODE* clone (CMSNODE *src, MEMSYS *mem)
{                               /* --- clone a c/m prefix subtree */
  CMSNODE *dst;                 /* created clone of the source node */
  CMSNODE **end;                /* end of the copied child list */
  CMSNODE *child;               /* to traverse the child nodes */

  assert(src && mem);           /* check the function arguments */
  dst = (CMSNODE*)ms_alloc(mem);
  if (!dst) return NULL;        /* create a clone of the node */
  dst->supp   = src->supp;      /* copy the sequence support, */
  dst->item   = src->item;      /* the associated item, */
  dst->height = src->height;    /* and the subtree height */
  end = &dst->children;         /* get destination child node list */
  for (child = src->children; child; child = child->sibling) {
    *end = clone(child, mem);   /* traverse and clone the children */
    if (!*end) { delete(dst, mem); return NULL; }
    end = &(*end)->sibling;     /* on error delete created clone */
  }                             /* advance the child node list end */
  *end = NULL;                  /* terminate the copied child list */
  return dst;                   /* return the created clone */
}  /* clone() */

/*--------------------------------------------------------------------*/

static int merge_pos (CMSNODE *dst, CMSNODE *src, MEMSYS *mem)
{                               /* --- merge a subtree to another */
  CMSNODE **end;                /* end of child node output list */
  CMSNODE *s, *d;               /* to traverse the child node lists */

  assert(dst && src && mem);    /* check the function arguments */
  end = &dst->children;         /* get the output child node list */
  s   = src->children;          /* get source and destination */
  d   = dst->children;          /* lists of child nodes */
  while (s && d) {              /* while both lists are not empty */
    if      (d->item < s->item){/* if an item is only in destination, */
      *end = d;                 /* simply transfer its subtree and */
      d = d->sibling; }         /* go to the next dest. child node */
    else if (d->item > s->item){/* if an item is only in source, */
      *end = clone(s, mem);     /* copy its subtree to destination */
      if (!*end) { *end = d; return -1; }
      s = s->sibling; }         /* go to the next source child node */
    else {                      /* if the item is in both lists */
      if (merge_pos(d, s, mem) < 0) {
        *end = d; return -1; }  /* recursively merge the subtrees */
      *end = d;                 /* transfer result to the dest. */
      d = d->sibling;           /* go to the next child nodes */
      s = s->sibling;           /* if both source and destination */
    }
    end = &(*end)->sibling;     /* advance the output position */
  }
  for ( ; s; s = s->sibling) {  /* traverse remaining source nodes */
    *end = clone(s, mem);       /* copy source subtree to destination */
    if (!*end) return -1;       /* (on error list is terminated) */
    end = &(*end)->sibling;     /* advance the output position */
  }
  *end = d;                     /* append remaining destination nodes */
  if (dst->height < src->height) dst->height = src->height;
  if (dst->supp   < src->supp)   dst->supp   = src->supp;
  return 0;                     /* return 'ok' */
}  /* merge_pos() */

/*--------------------------------------------------------------------*/

static int merge_neg (CMSNODE *dst, CMSNODE *src, MEMSYS *mem)
{                               /* --- merge a subtree to another */
  CMSNODE **end;                /* end of child node output list */
  CMSNODE *s, *d;               /* to traverse the child node lists */

  assert(dst && src && mem);    /* check the function arguments */
  end = &dst->children;         /* get the output child node list */
  s   = src->children;          /* get source and destination */
  d   = dst->children;          /* lists of child nodes */
  while (s && d) {              /* while both lists are not empty */
    if      (d->item > s->item){/* if an item is only in destination, */
      *end = d;                 /* simply transfer its subtree */
      d = d->sibling; }         /* go to the next dest. child node */
    else if (d->item < s->item){/* if an item is only in source, */
      *end = clone(s, mem);     /* copy its subtree to destination */
      if (!*end) { *end = d; return -1; }
      s = s->sibling; }         /* go to the next source child node */
    else {                      /* if the item is in both lists */
      if (merge_neg(d, s, mem) < 0) {
        *end = d; return -1; }  /* recursively merge the subtrees */
      *end = d;                 /* transfer result to the dest. */
      d = d->sibling;           /* go to the next child nodes */
      s = s->sibling;           /* if both source and destination */
    }
    end = &(*end)->sibling;     /* advance the output position */
  }
  for ( ; s; s = s->sibling) {  /* traverse remaining source nodes */
    *end = clone(s, mem);       /* copy source subtree to destination */
    if (!*end) return -1;       /* (on error list is terminated) */
    end = &(*end)->sibling;     /* advance the output position */
  }
  *end = d;                     /* append remaining destination nodes */
  if (dst->height < src->height) dst->height = src->height;
  if (dst->supp   < src->supp)   dst->supp   = src->supp;
  return 0;                     /* return 'ok' */
}  /* merge_neg() */

/*--------------------------------------------------------------------*/

static int project (CMSTREE *dst, CMSNODE *src)
{                               /* --- project a c/m prefix subtree */
  int     r;                    /* result of function call */
  CMSNODE *node;                /* to traverse the child nodes */

  assert(dst && src);           /* check the function arguments */
  for (node = src->children; node; node = node->sibling) {
    if (node->item != dst->item)/* if node item is different, */
      r = project(dst, node);   /* project the subtree */
    else if (dst->dir < 0)      /* if node item has projection item */
      r = merge_neg(&dst->root, node, dst->mem);
    else                        /* merge source subtree to dest. */
      r = merge_pos(&dst->root, node, dst->mem);
    if (r < 0) return r;        /* (respect the item order direction) */
  }                             /* abort if an error occurs */
  return 0;                     /* return 'ok' */
}  /* project() */

/*--------------------------------------------------------------------*/

CMSTREE* cms_project (CMSTREE *dst, CMSTREE *src, ITEM item)
{                               /* --- project a c/m prefix tree */
  CMSTREE *arg = dst;           /* note the destination argument */

  assert(src && (item >= 0)     /* check the function arguments */
  &&    (!dst || (dst->dir == src->dir)));
  if (!dst) dst = cms_create(src->size, src->target, src->dir);
  if (!dst) return NULL;        /* create/reinit. destination tree */
  dst->item = item;             /* note the projection item */
  dst->root.supp = 0;           /* clear maximum item set support */
  assert(!dst->root.children);  /* check for an empty destination */
  if (project(dst, &src->root) < 0) {  /* project source to dest. */
    if (!arg) { cms_delete(dst, 1); } return NULL; }
  return dst;                   /* return the destination tree */
}  /* cms_project() */

/*--------------------------------------------------------------------*/

static RSUPP clo_perm (CMSTREE *cms, CMSNODE *node,
                       ITEM pos, RSUPP supp)
{                               /* --- prune non-closed permutations */
  CMSNODE **p;                  /* to traverse the sibling list */
  RSUPP   m = -1;               /* maximal support of a sibling node */
  RSUPP   s;                    /* maximal support of descendants */
  ITEM    k;                    /* permutation position of an item */

  assert(cms                    /* check the function arguments */
  &&    node && (pos >= -1) && (pos < cms->size) && (supp >= 0));
  p = &node->children;          /* get the child node list and */
  while (*p) {                  /* traverse the sibling nodes */
    node = *p;                  /* get the next sibling node */
    k = cms->pos[node->item];   /* and the position of its item */
    if (k <= pos) {             /* if node represents no sub-perm. */
      p = &node->sibling;       /* skip node, update maximal support */
      if (m < node->supp) m = node->supp; }
    else {                      /* if node represents a sub-perm. */
      s = clo_perm(cms, node, k, supp);
      if (node->supp > supp) { p = &node->sibling; s = node->supp; }
      else if (s >= 0)       { p = &node->sibling; node->supp = s; }
      else { *p = node->sibling; ms_free(cms->mem, node); }
      if (m < s) m = s;         /* recursively prune child nodes */
    }                           /* and update the maximal support */
  }                             /* of a sibling node */
  return m;                     /* return the maximal sibling support */
}  /* clo_perm() */

/*--------------------------------------------------------------------*/

static RSUPP max_perm (CMSTREE *cms, CMSNODE *node, ITEM pos)
{                               /* --- prune non-maximal permutations */
  CMSNODE **p;                  /* to traverse the sibling list */
  RSUPP   m = -1;               /* maximal support of a sibling node */
  RSUPP   s;                    /* maximal support of descendants */
  ITEM    k;                    /* permutation position of an item */

  assert(cms                    /* check the function arguments */
  &&     node && (pos >= -1) && (pos < cms->size));
  p = &node->children;          /* get the child node list and */
  while (*p) {                  /* traverse the sibling nodes */
    node = *p;                  /* get the next sibling node */
    k = cms->pos[node->item];   /* and the position of its item */
    if (k <= pos) {             /* if node represents no sub-perm. */
      p = &node->sibling;       /* skip node, update maximal support */
      if (node->supp > m) m = node->supp; }
    else {                      /* if node represents a sub-perm. */
      s = max_perm(cms, node, k);
      if (s < 0) { *p = node->sibling; ms_free(cms->mem, node); }
      else { p = &node->sibling; if (m < (node->supp = s)) m = s; }
    }                           /* recursively prune child nodes */
  }                             /* and update the maximal support */
  return m;                     /* return the maximal sibling support */
}  /* max_perm() */

/*--------------------------------------------------------------------*/

static RSUPP clo_seq (CMSTREE *cms, CMSNODE *node,
                      const ITEM *items, ITEM n, RSUPP supp)
{                               /* --- prune non-closed sequences */
  CMSNODE **p;                  /* to traverse the sibling list */
  RSUPP   m = -1;               /* maximal support of a sibling node */
  RSUPP   s;                    /* maximal support of descendants */
  ITEM    k;                    /* loop variable for item positions */

  assert(cms                    /* check the function arguments */
  &&     node && items && (n >= 0) && (supp >= 0));
  p = &node->children;          /* get the child node list and */
  while (*p) {                  /* traverse the sibling nodes */
    node = *p;                  /* get the next sibling node */
    if (cms->pos[node->item] < 0)
      k = n;                    /* if node represents no sub-sequence */
    else {                      /* if node may represent a sub-seq., */
      for (k = 0; k < n; k++)   /* traverse the item sequence */
        if (items[k] == node->item)
          break;                /* find item in sequence (suffix) */
    }                           /* (check path for a subsequence) */
    if (k >= n) {               /* if the item could not be found, */
      p = &node->sibling;       /* skip node & update maximal support */
      if (node->supp > m) m = node->supp; }
    else {                      /* if item could be found */
      s = clo_seq(cms, node, items+k+1, n-(k+1), supp);
      if (node->supp > supp) { p = &node->sibling; s = node->supp; }
      else if (s >= 0)       { p = &node->sibling; node->supp = s; }
      else { *p = node->sibling; ms_free(cms->mem, node); }
      if (m < s) m = s;         /* recursively prune child nodes */
    }                           /* and update the maximal support */
  }
  return m;                     /* return the maximal sibling support */
}  /* clo_seq() */

/*--------------------------------------------------------------------*/

static RSUPP max_seq (CMSTREE *cms, CMSNODE *node,
                      const ITEM *items, ITEM n)
{                               /* --- prune non-maximal sequences */
  CMSNODE **p;                  /* to traverse the sibling list */
  RSUPP   m = -1;               /* maximal support of a sibling node */
  RSUPP   s;                    /* maximal support of descendants */
  ITEM    k;                    /* loop variable for item positions */

  assert(cms                    /* check the function arguments */
  &&     node && items && (n >= 0));
  p = &node->children;          /* get the child node list and */
  while (*p) {                  /* traverse the sibling nodes */
    node = *p;                  /* get the next sibling node */
    if (cms->pos[node->item] < 0)
      k = n;                    /* if node represents no sub-sequence */
    else {                      /* if node may represent a sub-seq., */
      for (k = 0; k < n; k++)   /* traverse the item sequence */
        if (items[k] == node->item)
          break;                /* find item in sequence (suffix) */
    }                           /* (check path for a subsequence) */
    if (k >= n) {               /* if the item could not be found, */
      p = &node->sibling;       /* skip node & update maximal support */
      if (node->supp > m) m = node->supp; }
    else {                      /* if item could be found */
      s = max_seq(cms, node, items+k+1, n-(k+1));
      if (s < 0) { *p = node->sibling; ms_free(cms->mem, node); }
      else { p = &node->sibling; if (m < (node->supp = s)) m = s; }
    }                           /* recursively prune child nodes */
  }                             /* and update the maximal support */
  return m;                     /* return the maximal sibling support */
}  /* max_seq() */

/*--------------------------------------------------------------------*/

void cms_prune (CMSTREE *cms, const ITEM *items, ITEM n, RSUPP supp)
{                               /* --- prune subsumed sequences */
  ITEM i;                       /* loop variable for item positions */

  assert(cms                    /* check the function arguments */
  &&    (items || (n <= 0)) && (supp > 0));
  for (i = 0; i < n; i++)       /* traverse the given item sequence */
    cms->pos[items[i]] = i;     /* and set the (last) item positions */
  if (cms->target & CMS_PERMUT){/* if item permutations (no reps.) */
    if (cms->target & CMS_MAXIMAL)
         max_perm(cms, &cms->root, -1);
    else clo_perm(cms, &cms->root, -1, supp); }
  else {                        /* if general sequences (reps.) */
    if (cms->target & CMS_MAXIMAL)
         max_seq (cms, &cms->root, items, n);
    else clo_seq (cms, &cms->root, items, n, supp);
  }
  for (i = 0; i < n; i++)       /* traverse the item sequence */
    cms->pos[items[i]] = -1;    /* and clear the item positions */
}  /* cms_prune() */

/*--------------------------------------------------------------------*/

int cms_update (CMSTREE *cms, const ITEM *items, ITEM n, RSUPP supp)
{                               /* --- update a c/m prm. prefix tree */
  RSUPP smax;                   /* max. extension support for target */

  assert(cms                    /* check the function arguments */
  &&    (items || (n <= 0)) && (supp > 0));
  smax = (cms->target & CMS_MAXIMAL) ? 1 : supp;
  if (cms_super(cms, items, n, smax) != 0)
    return 1;                   /* check for a super-sequence */
  cms_prune(cms, items, n, supp);
  return cms_add(cms, items, n, supp);
}  /* cms_update() */           /* prune sub-seqs. and add sequence */

/*--------------------------------------------------------------------*/

static int closed (CMSTREE *cms, CMSNODE *node)
{                               /* --- report closed item sequence */
  int  r, x = 0;                /* error status, perfect ext. flag */
  SUPP supp;                    /* support of current item sequence */

  assert(cms && node);          /* check the function arguments */
  supp = isr_supp(cms->rep);    /* get current item sequence support */
  if (isr_xable(cms->rep, 1)) { /* if item sequence may be extended */
    for (node = node->children; node; node = node->sibling) {
      x |= (node->supp >= supp);/* set perfect extension flag */
      r  = isr_addnc(cms->rep, node->item, node->supp);
      if (r < 0) return r;      /* add current item to the reporter */
      r  = closed(cms, node);   /* recursively report item sequences */
      isr_remove(cms->rep, 1);  /* and remove the current item */
      if (r < 0) return r;      /* from the item set reporter, */
    } }                         /* finally check for an error */
  else {                        /* if sequence may not be extended */
    for (node = node->children; node; node = node->sibling)
      if (node->supp >= supp) { x = -1; break; }
  }                             /* check for a perfect extension */
  if (!x && (isr_report(cms->rep) < 0))
    return -1;                  /* report sequences w/ perfect exts. */
  return 0;                     /* return 'ok' */
}  /* closed() */

/*--------------------------------------------------------------------*/

static int maximal (CMSTREE *cms, CMSNODE *node)
{                               /* --- report maximal item sequences */
  int r, x = 0;                 /* error status, freq. extension flag */

  assert(cms && node);          /* check the function arguments */
  x = (node->children != NULL); /* note whether there are children */
  if (isr_xable(cms->rep, 1)) { /* if item sequence may be extended */
    for (node = node->children; node; node = node->sibling) {
      r = isr_addnc(cms->rep, node->item, node->supp);
      if (r < 0) return r;      /* add current item to the reporter */
      r = maximal(cms, node);   /* recursively report item sequences */
      isr_remove(cms->rep, 1);  /* and remove the current item */
      if (r < 0) return r;      /* from the item set reporter, */
    }                           /* finally check for an error */
  }
  if (!x && (isr_report(cms->rep) < 0))
    return -1;                  /* report seqs. without super-seqs. */
  return 0;                     /* return 'ok' */
}  /* maximal() */

/*--------------------------------------------------------------------*/

int cms_report (CMSTREE *cms, ISREPORT *rep)
{                               /* --- report c/m item sequences */
  assert(cms && rep);           /* check the function arguments */
  cms->rep = rep;               /* note the item set reporter */
  if (cms->target & CMS_MAXIMAL) return maximal(cms, &cms->root);
  else                           return closed (cms, &cms->root);
}  /* cms_report() */           /* recursively report sequences */

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

static void indent (int k)
{ while (--k >= 0) printf("   "); }

/*--------------------------------------------------------------------*/

static void show (CMSNODE *node, ITEMBASE *base, int ind)
{                               /* --- recursively show nodes */
  assert(ind >= 0);             /* check the function arguments */
  indent(ind);                  /* indent the output line */
  if (base && (node->item >= 0))/* print the associated item */
    printf("%s/", ib_xname(base, node->item));
  printf("%"ITEM_FMT":",  node->item);   /* print node information */
  printf("%"SUPP_FMT":",  node->supp);   /* (item identifier, */
  printf("%"ITEM_FMT"\n", node->height); /* support and height) */
  for (node = node->children; node; node = node->sibling)
    show(node, base, ind+1);    /* recursively show the child nodes */
}  /* show() */

/*--------------------------------------------------------------------*/

void cms_show (CMSTREE *cms, ITEMBASE *base, int ind)
{                               /* --- print a c/m prefix tree */
  indent(ind);                  /* indent the output line */
  if (!cms) {                   /* check whether tree exists */
    printf("(null)\n"); return; }
  printf("item: %d", cms->item);/* print associated item identifier */
  if (base && (cms->item >= 0)) /* print item name if possible */
    printf("/%s", ib_name(base, cms->item));
  printf("\n");                 /* terminate the item id/name */
  show(&cms->root, base, ind);  /* recursively show the nodes */
}  /* cms_show() */

#endif
/*----------------------------------------------------------------------
  Closed/Maximal Filter Functions
----------------------------------------------------------------------*/

CM4SEQS* cmx_create (ITEM size, ITEM zmax, int target, int dir)
{                               /* --- create a c/m filter object */
  CM4SEQS *cmx;                 /* created closed/maximal filter */
  CMSTREE *t;                   /* created root prefix tree */

  assert(size > 0);             /* check the function arguments */
  cmx = (CM4SEQS*)calloc(1, sizeof(CM4SEQS)
                         +(size_t)zmax *sizeof(CMSTREE*));
  if (!cmx) return NULL;        /* create a closed/maximal filter */
  cmx->target = target;         /* and initialize its fields */
  cmx->dir    = dir;
  cmx->size   = size;
  cmx->cnt    = 1;              /* create and init. a root tree */
  cmx->trees[0] = t = cms_create(size, target, dir);
  if (!t) { cmx_delete(cmx); return NULL; }
  return cmx;                   /* return the created c/m filter */
}  /* cmx_create() */

/*--------------------------------------------------------------------*/

void cmx_delete (CM4SEQS *cmx)
{                               /* --- delete a c/m filter object */
  ITEM i;                       /* loop variable */

  assert(cmx);                  /* check the function argument */
  for (i = 0; cmx->trees[i]; i++)
    cms_delete(cmx->trees[i],1);/* delete the c/m prefix trees */
  free(cmx);                    /* delete the base structure */
}  /* cmx_delete() */

/*--------------------------------------------------------------------*/

RSUPP cmx_supp (CM4SEQS *cmx)
{                               /* --- get support of current prefix */
  assert(cmx);                  /* check the function argument */
  return cms_supp(cmx->trees[(cmx->cnt > 0) ? cmx->cnt-1 : 0]);
}  /* cmx_supp() */

/*--------------------------------------------------------------------*/

int cmx_add (CM4SEQS *cmx, ITEM item, RSUPP supp)
{                               /* --- add an item to the prefix */
  CMSTREE *t, **p;              /* to access the c/m prefix trees */

  assert(cmx                    /* check the function arguments */
  &&    (item >= 0) && (item < cmx->size));
  p = cmx->trees +cmx->cnt;     /* get the current prefix tree */
  t = cms_project(*p, p[-1], item);
  if (!t) return -1;            /* project tree to the given item */
  *p = t;                       /* and store the created projection */
  ++cmx->cnt; return 0;         /* count the added item */
}  /* cmx_add() */

/*--------------------------------------------------------------------*/

void cmx_remove (CM4SEQS *cmx, ITEM n)
{                               /* --- remove items from the prefix */
  assert(cmx && (n >= 0));      /* check the function arguments */
  for (n = (n < cmx->cnt) ? cmx->cnt -n : 1; cmx->cnt > n; )
    if (cmx->trees[--cmx->cnt]) cms_clear(cmx->trees[cmx->cnt]);
}  /* cmx_remove() */           /* traverse and clear the trees */

/*--------------------------------------------------------------------*/

int cmx_update (CM4SEQS *cmx, const ITEM *items, ITEM n, RSUPP supp)
{                               /* --- update filter with found seq. */
  int  r;                       /* result of cms_update() */
  ITEM i;                       /* loop variable */

  assert(cmx                    /* check function arguments */
  &&    (items || (n <= 0)) && (supp >= 0));
  for (i = (cmx->cnt < n+1) ? cmx->cnt : n+1; --i >= 0; ) {
    r = cms_update(cmx->trees[i], items+i, n-i, supp);
    if (r < 0) return -1;       /* traverse the prefix trees and */
    if (r > 0) return +1;       /* add the item sequence to all trees */
  }                             /* (unless there is a super-sequence) */
  return 0;                     /* return 'ok' (update was done) */
}  /* cmx_update() */

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

void cmx_show (CM4SEQS *cmx, ITEMBASE *base, int ind)
{                               /* --- print a closed/maximal filter */
  ITEM i;                       /* loop variable */

  assert(cmx);                  /* check the function arguments */
  for (i = 0; i <= cmx->cnt; i++) /* traverse and print the c/m trees */
    cms_show(cmx->trees[i], base, ind);
}  /* cmx_show() */

#endif
/*----------------------------------------------------------------------
  Main Function for Testing
----------------------------------------------------------------------*/
#ifdef CMS_MAIN

#define SIZE(a)    (sizeof(a)/sizeof(*a))

int main (void)
{
  CMSTREE *cms, *dst;
  ITEM    a[] = { 1, 3, 5 };             RSUPP sa = 8;
  ITEM    b[] = { 2, 4, 6 };             RSUPP sb = 7;
  ITEM    c[] = { 1, 2, 3, 4, 5, 6 };    RSUPP sc = 5;
  ITEM    d[] = { 0, 1, 2, 3, 4, 5, 6 }; RSUPP sd = 5;
  ITEM    e[] = { 1, 2, 4, 6 };          RSUPP se = 6;
  int     max, m;

  for (max = 1; max >= 0; max--) {
    printf((max) ? "maximal\n" : "closed\n");
    m = (max) ? CMS_MAXIMAL : CMS_CLOSED;
    cms = cms_create(10, CMS_PERMUT|m, -1);
    if (!cms) return -1;
    cms_show(cms, NULL, 0);
    printf("\n");

    if (cms_update(cms, a, SIZE(a), sa) < 0) return -1;
    cms_show(cms, NULL, 0);
    printf("\n");

    if (cms_update(cms, b, SIZE(b), sb) < 0) return -1;
    cms_show(cms, NULL, 0);
    printf("\n");

    if (cms_update(cms, c, SIZE(c), sc) < 0) return -1;
    cms_show(cms, NULL, 0);
    printf("\n");

    if (cms_update(cms, d, SIZE(d), sd) < 0) return -1;
    cms_show(cms, NULL, 0);
    printf("\n");

    if (cms_update(cms, e, SIZE(e), se) < 0) return -1;
    cms_show(cms, NULL, 0);
    printf("\n");

    printf("projection to 2\n");
    dst = cms_project(NULL, cms, 2);
    cms_show(dst, NULL, 0);
    printf("\n");

    cms_delete(dst, 1);
    cms_delete(cms, 1);
  }

  return 0;
}  /* main() */

#endif
